from importlib.util import module_from_spec, spec_from_file_location
import logging

import click

from .proxy import Proxy


log = logging.getLogger(__name__)


def configure_logging(level):
    logging.basicConfig(format='[%(levelname)s] :: %(message)s', level=level)


def load_script(path):
    spec = spec_from_file_location('script', path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    log.info('Loaded script %s.', path)


@click.group()
def cli():
    pass


@cli.command()
@click.option('--debug', '-d', default=False, is_flag=True)
@click.option('--port', '-p', default=8080)
@click.option('--script', '-s', multiple=True, type=click.Path(exists=True))
def run(debug, port, script):

    configure_logging(logging.INFO if debug else logging.WARNING)

    # load scripts before instantiating the proxy to allow the use of the `on` decorator
    for path in script:
        load_script(path)

    proxy = Proxy(host='0.0.0.0', port=port)
    proxy.run()



import logging

from mitmproxy.controller import handler as flow_handler
from mitmproxy.flow import FlowMaster, State
from mitmproxy.models import decoded
from mitmproxy.options import Options
from mitmproxy.proxy import ProxyConfig, ProxyServer

from .constants import ALL, GUNGHO_API_ENDPOINT, GUNGHO_USER_AGENT
from .structures import CaseInsensitiveDefaultDict


log = logging.getLogger(__name__)


class BaseProxy(FlowMaster):


    def __init__(self, **options):
        opts = Options(**options)
        server = ProxyServer(ProxyConfig(opts))
        super().__init__(opts, server, State())


    def run(self):
        try:
            logging.info('Starting proxy on port %s.', self.options.listen_port)
            super().run()
        except KeyboardInterrupt:
            pass
        finally:
            logging.info('Shutting down.')
            self.shutdown()


class Proxy(BaseProxy):

    handlers = CaseInsensitiveDefaultDict(set)

    def __init__(self, host='0.0.0.0', port=8080):
        super().__init__(listen_host=host, listen_port=port, mode='transparent')
        self.handlers = Proxy.handlers.copy()


    @flow_handler
    def response(self, flow):
        request = flow.request

        log.debug('Received response from %s request to %s.', request.method, request.pretty_url)

        if request.headers.get('user-agent') == GUNGHO_USER_AGENT and request.path.startswith(GUNGHO_API_ENDPOINT):
            log.info('Captured %s request to %s. Forwarding flow to routing.', request.method, request.pretty_url)
            self.route(flow)


    def on(self, action):
        """
        Register a function to be called when a response is received from a request to `action`.

        The function will be called with the request and response. The response will not be received by the client
        until the decorated function returns.

        Read about possible actions on the padsniff wiki: https://bitbucket.org/necromanteion/padsniff/wiki/Home
        """
        def wrapper(func):
            self.handlers[action].add(func)
            return func

        return wrapper


    def route(self, flow):
        """
        Route a flow's request and response to a handler function based on its action URL parameter.
        """
        request, response = flow.request, flow.response
        action = request.query.get('action')
        funcs = self.handlers[action] | self.handlers[ALL]

        for func in funcs:
            try:
                func(request, response)
            except:
                log.exception('Error while executing %s.', func.__name__)


def on(action, *, cls=Proxy):
    return cls.on(cls, action)



from collections.abc import MutableMapping


class CaseInsensitiveDefaultDict(MutableMapping):
    """
    A mashup of a case-insensitive keyed dict and collections.defaultdict.

    Write an example for me!
    """


    @staticmethod
    def _transform(key):
        """Supports non-string keys."""

        try:
            return key.casefold()

        except AttributeError:
            return key


    def __init__(self, default_factory, data=None, **kwargs):
        self._dict = {}

        self.default_factory = default_factory

        if data is None:
            data = {}

        self.update(data, **kwargs)


    def __setitem__(self, key, value):
        self._dict[self._transform(key)] = (key, value)


    def __getitem__(self, key):
        try:
            origkey, value = self._dict[self._transform(key)]

        except KeyError:
            # call to __missing__ must be explicit because of overloading __getitem__
            value = self.__missing__(key)

        return value


    def __delitem__(self, key):
        del self._dict[self._transform(key)]


    def __iter__(self):
        return (origkey for origkey, value in self._dict.values())


    def __len__(self):
        return len(self._dict)


    def __missing__(self, key):
        # emulate defaultdict's behavior is no default_factory is supplied
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = self.default_factory()
        return self[key]


    def __repr__(self):
        line = '{0.__class__.__qualname__}('
        line += (repr(None) if self.default_factory is None
                 else '{0.default_factory.__qualname__}')
        line += ', {1})'
        return line.format(self, dict(self.items()))


    def copy(self):
        return CaseInsensitiveDefaultDict(self.default_factory, self._dict.values())


    def get(self, key, default=None):

        try:
            # _dict.get avoids __getitem__'s defaultdict-like behavior
            origkey, value = self._dict.get(self._transform(key))

        except TypeError:
            value = default

        return value


    @classmethod
    def fromkeys(cls, default_factory, seq, value=None):
        return cls(default_factory, dict.fromkeys(seq, value))